from pwn import *
import marshal
import dis
import sys
from collections import defaultdict

assert sys.version.startswith('3.11.')

context.endian = 'little'

with open('./x.pyc', 'rb') as f:
    pyc = f.read()

mod = marshal.loads(pyc[16:])
code = type(mod)
funcs = {
    x.co_name: x
    for x in mod.co_consts
    if isinstance(x, code)
}
funcs[mod.co_name] = mod
funcs['w'] = funcs['game1'].co_consts[1]

func_starts = {
    '<module>': 0x2a,
    'ks': 0x94,
    'cry': 0x20e,
    'fail': 0x370,
    'game1': 0x489,
    'w': 0x7a2,
    'game2': 0xc0c,
    'game3': 0x105f,
    'main': 0x1654,
}

binary_ops = {
    0: 'ADD',
    1: 'AND',
    2: 'FLOOR_DIVIDE',
    3: 'LSHIFT',
    4: 'MATRIX_MULTIPLY',
    5: 'MULTIPLY',
    6: 'REMAINDER',
    7: 'OR',
    8: 'POWER',
    9: 'RSHIFT',
    10: 'SUBSTRACT',
    11: 'TRUE_DIVIDE',
    12: 'XOR',
    13: 'INPLACE_ADD',
    14: 'INPLACE_AND',
    15: 'INPLACE_FLOOR_DIVIDE',
    16: 'INPLACE_LSHIFT',
    17: 'INPLACE_MATRIX_MULTIPLY',
    18: 'INPLACE_MULTIPLY',
    19: 'INPLACE_REMAINDER',
    20: 'INPLACE_OR',
    21: 'INPLACE_POWER',
    22: 'INPLACE_RSHIFT',
    23: 'INPLACE_SUBTRACT',
    24: 'INPLACE_TRUE_DIVIDE',
    25: 'INPLACE_XOR'
}

cmp_op = ('<', '<=', '==', '!=', '>', '>=')

known_ops = {
    # opnum: (name, hasArg)
    96: ('LOAD_CONST', True),
    171: ('IMPORT_NAME', True),
    150: ('STORE_NAME', True),
    99: ('MAKE_FUNCTION', True),
    49: ('RETURN_VALUE', True),
    97: ('RESUME', True),
    185: ('LOAD_GLOBAL', True),
    188: ('LOAD_ATTR', True),
    191: ('PRECALL', True),
    142: ('CALL', True),
    1: ('POP_TOP', False),
    198: ('BINARY_OP', True),
    98: ('LOAD_FAST', True),
    141: ('STORE_FAST', True),
    146: ('LOAD_METHOD', True),
    50: ('<<MARKER>>', True),
    104: ('COMPARE_OP', True),
    175: ('POP_JUMP_BACKWARD_IF_TRUE', True),
    186: ('POP_JUMP_FORWARD_IF_FALSE', True),
    94: ('JUMP_FORWARD', True),
    123: ('JUMP_BACKWARD', True),
    82: ('BINARY_SUBSCR', False),
    143: ('UNPACK_SEQUENCE', True),
    69: ('YIELD', False),
    130: ('BUILD_TUPLE', True),
    117: ('CONTAINS_OP', True),
    31: ('GET_ITER', False),
    182: ('FOR_ITER', True),
    116: ('BUILD_LIST', True),
    27: ('NOP', False),
    177: ('BUILD_CONST_KEY_MAP', True),
    138: ('BUILD_STRING', True),
    136: ('BUILD_SLICE', True),
}

unknown_opcodes_seen = defaultdict(int)

for funcname, funcstart in func_starts.items():
    f = funcs[funcname]
    print(f'{funcname}({",".join(f.co_varnames[:f.co_argcount])}):')
    print('\n'.join('\t' + x for x in dis._format_code_info(f).split('\n')))
    line_starts = {line:offset for line,offset in dis.findlinestarts(f)}
    codelen = u32(pyc[funcstart - 4:][:4])
    code = pyc[funcstart:][:codelen]
    nops = 0
    idx = 0
    ismarker, markercount = False, 0
    while idx < len(code):
        if idx in line_starts:
            print(f' {line_starts[idx]}')
        op = u8(code[idx:][:1])
        arg = u8(code[idx+1:][:1])
        if op == 32:
            assert arg == 0
            nops += 1
            idx += 2
            continue
        elif nops > 0:
            # print(f'  ...{nops} nops...')
            nops = 0
        marker = f'>>{markercount:02}<<' if ismarker else '      '
        markercount += 1 if ismarker else 0
        ismarker = False
        if op in known_ops:
            opname, hasargs = known_ops[op]
            if opname == '<<MARKER>>':
                assert arg == 0
                ismarker = True
            elif opname == 'CALL':
                print(f'    {idx:>04x}: {op:>02x} {arg:>02x}  {marker} {opname} with {arg} arguments')
            elif opname == 'PRECALL':
                print(f'    {idx:>04x}: {op:>02x} {arg:>02x}  {marker} {opname} with {arg} arguments')
            elif opname == 'LOAD_GLOBAL':
                assert arg % 2 == 0
                print(f'    {idx:>04x}: {op:>02x} {arg:>02x}  {marker} {opname} {f.co_names[arg//2]!r}')
            elif opname == 'LOAD_CONST':
                print(f'    {idx:>04x}: {op:>02x} {arg:>02x}  {marker} {opname} {f.co_consts[arg]!r}')
            elif opname in ['LOAD_ATTR', 'STORE_NAME', 'LOAD_METHOD', 'IMPORT_NAME']:
                print(f'    {idx:>04x}: {op:>02x} {arg:>02x}  {marker} {opname} {f.co_names[arg]!r}')
            elif opname in ['LOAD_FAST', 'STORE_FAST']:
                vartyp = 'arg' if arg < f.co_argcount else 'var'
                print(f'    {idx:>04x}: {op:>02x} {arg:>02x}  {marker} {opname} {vartyp} {f.co_varnames[arg]!r}')
            elif opname == 'BINARY_OP':
                binop = binary_ops[arg] if arg in binary_ops else f'????? {arg}'
                print(f'    {idx:>04x}: {op:>02x} {arg:>02x}  {marker} {opname} {binop}')
            elif opname == 'COMPARE_OP':
                print(f'    {idx:>04x}: {op:>02x} {arg:>02x}  {marker} {opname} {cmp_op[arg]}')
            elif opname == 'CONTAINS_OP':
                print(f'    {idx:>04x}: {op:>02x} {arg:>02x}  {marker} {opname} {["in","not in"][arg]}')
            elif opname == 'RESUME':
                resumetype = {
                    0: 'func start',
                    1: 'after yield',
                    2: 'after yield from',
                    3: 'after await',
                }[arg]
                print(f'    {idx:>04x}: {op:>02x} {arg:>02x}  {marker} {opname} {resumetype}')
            elif opname in ['POP_JUMP_BACKWARD_IF_TRUE', 'POP_JUMP_FORWARD_IF_FALSE',
                            'JUMP_FORWARD', 'JUMP_BACKWARD', 'FOR_ITER']:
                signed_arg = -arg if 'BACKWARD' in opname else arg
                tgt = idx + 2 + signed_arg*2
                tgtline = line_starts[tgt] if tgt in line_starts else '??'
                prop = 'if exhaused goto' if opname == 'FOR_ITER' else 'to'
                print(f'    {idx:>04x}: {op:>02x} {arg:>02x}  {marker} {opname} {prop} {tgt:>04x} (line {tgtline})')
            elif hasargs:
                print(f'    {idx:>04x}: {op:>02x} {arg:>02x}  {marker} {opname} {arg}')
            else:
                if arg == 0:
                    print(f'    {idx:>04x}: {op:>02x} {arg:>02x}  {marker} {opname}')
                else:
                    print(f'  WEIRD: {opname} unexpected arg {arg:>02x} ({funcname} : {hex(funcstart+idx)})')
        else:
            unknown_opcodes_seen[op] += 1
            print(f'    {idx:>04x}: {op:>02x} {arg:>02x}  {marker} <{op}> <{arg}>')
        idx += 2
    print()

print()
print("Unknown opcodes seen:")
for count, opcode in sorted([(v, k) for k, v in unknown_opcodes_seen.items()])[::-1]:
    print(count, opcode)
